use bitboard::Bitboard;
use board::Board;
use search::*;
use movegen::*;
use std::process::exit;
use std::sync::mpsc::{Sender, self};
use std::time::Duration;
use zobrist::ZobristTable;
use std::thread;

use std::sync::{Arc, RwLock};
use crate::ttable::{RepetitionTable, Cache};
use crate::uci;

///
/// Message types that are sent from the uci interface to the engine thread.
/// 
pub enum Command {
    Uci,
    Position{ pos: Board, moves: Vec<String> },
    SetPiece(Bitboard),
    Go(SearchInfo),
    Print,
    IsReady,
    Ping,
    Stop,
    NewGame,
    GetState,
    GetInfo,
    Quit,
}

///
/// Structure holding information for a search command.
/// 
pub struct SearchInfo {
    pub depth: Option<i32>,
    pub movetime: Option<i32>,
    pub wtime: Option<isize>,
    pub btime: Option<isize>,
    pub winc: Option<isize>,
    pub binc: Option<isize>,
    pub movestogo: Option<isize>,
    pub perft: bool,
    pub infinite: bool
}

#[derive(PartialEq)]
pub enum SearchMessage {
    Stop
}

///
/// Stores the state of the chess engine.
/// 
pub struct Engine {
    board: Board,
    move_history: RepetitionTable,
    hash: Arc<RwLock<Cache>>,
    zobrist_table: Arc<ZobristTable>,

    search_thread: Option<SearchThread>,
}

struct SearchThread {
    channel: Sender<SearchMessage>,
    join_handle: thread::JoinHandle<bool>,
}

impl SearchInfo {
    pub fn new() -> Self {
        SearchInfo {
            depth: None,
            movetime: None,
            wtime: None,
            btime: None,
            winc: None,
            binc: None,
            movestogo: None,
            perft: false,
            infinite: false
        }
    }
}

impl Engine {
    pub fn new() -> Self {
        let mut eng = Engine {
            board: Board::default(),
            move_history: RepetitionTable::new(),
            hash: Arc::new(RwLock::new(Cache::megabytes(4))),
            zobrist_table: Arc::new(ZobristTable::new()),
            search_thread: None,
        };
        eng.board.zobrist = Some((eng.zobrist_table.clone(), eng.board.calculate_zobrist(&eng.zobrist_table)));
        eng
    }

    pub fn run(&mut self, input: std::io::Stdin) {
        for line_m in input.lines() {
            let line = line_m.unwrap();

            //println!("received line: {}", line);

            let command = uci::parse_command(&line);

            if let Ok(cmd) = command {
                self.run_command(cmd);
            }
            else if let Err(err) = command {
                println!("Error parsing command: {}", err);
            }
        }
    }

    fn run_command(&mut self, cmd: Command) {
        match cmd {
            Command::Uci => {
                println!("id name bishop 1.0");
                println!("id author Geile Siech");
                println!("uciok");
            },
            Command::IsReady => {
                println!("readyok");
            },
            Command::Position { pos, moves } => {
                self.set_position(pos, &moves);
            },
            Command::Go(si) => {
                self.stop_search();
                self.start_search(&si);
            },
            Command::Stop => {
                self.stop_search();
            },
            Command::Quit => {
                exit(0);
            }
            _ => {
                println!("not yet implemented!");
            }
        }
    }


    fn set_position(&mut self, pos: Board, moves: &Vec<String>) {
        self.board = pos;

        self.board.zobrist = Some((self.zobrist_table.clone(), self.board.calculate_zobrist(&self.zobrist_table)));
        self.move_history.clear();
        self.move_history.increment(self.board.calculate_zobrist(&self.zobrist_table));
        for mov in moves {
            let m = self.board.parse_move(&mov);
            if let Ok(mm) = m {
                self.board.apply(mm);
                self.move_history.increment(self.board.calculate_zobrist(&self.zobrist_table));
            }
            else {
                println!("{}", self.board.beautiful_print());
                println!("error parsing move {}", mov);
                break;
            }
        }
    }

    fn start_search(&mut self, si: &SearchInfo) {
        let (snd, rcv) = mpsc::channel::<SearchMessage>();

        let join_handle = 
            if si.perft {
                let depth = si.depth.unwrap_or(0);
                let mut pc = SearchControl::new_perft(&self.board, rcv, depth);
                let search_rtn = move || pc.perft(depth);
                thread::spawn(search_rtn)
            }
            else {
                let mut sc = SearchControl::new(&self.board, rcv, self.hash.clone());
                sc.hash = self.hash.clone();
                sc.movetime = si.movetime.map(|ms| Duration::from_millis(ms as _));
                if self.board.turn == WHITE {
                    sc.remaining_time = si.wtime.map(|ms| Duration::from_millis(ms as _));
                }
                else if self.board.turn == BLACK {
                    sc.remaining_time = si.btime.map(|ms| Duration::from_millis(ms as _));
                }
                let depth = si.depth;
                let search_rtn = move || sc.iterative_deepening(depth);
                thread::spawn(search_rtn)
            };

        self.search_thread = Some(SearchThread {
            channel: snd,
            join_handle
        });
        //self.search_threads.push(st);
    }

    fn stop_search(&mut self) {
        if let Some(st) = std::mem::replace(&mut self.search_thread, None) {
            let _msg_res = st.channel.send(SearchMessage::Stop);
            let _res = st.join_handle.join().unwrap();
        }
    }

    fn reconstruct_pv(game: &Board, hash: &Cache) -> Vec<Move> {
        let mut pv: Vec<Move> = Vec::new();
        let mut g = game.clone();

        // max 100 to avoid infinite loops
        for _ in 0..100 {
            let mce = hash.lookup(&g);
            if let Some(ce) = mce {
                if ce.mov == (SimpleMove{ from: 0, to: 0 }) {
                    break;
                }
                let movs = generate_moves(&g, g.turn);
                let bm = movs.iter().find(|m| (**m).to_simple() == ce.mov);
                if let Some(found) = bm {
                    g.apply(*found);
                    pv.push(*found);
                }
                else {
                    break;
                }
            }
            else {
                break;
            }
        }
        return pv;
    }
}
