use std::io::{self, BufRead};
use std::collections::BTreeMap;
use std::sync::mpsc::{Receiver, Sender};

use std::process::exit;

use engine::{Command, SearchInfo};
use board::Board;

use log::info;

use crate::evaluate::PosValue;


pub fn parse_command(command: &str) -> Result<Command, String> {
    let parts = command.split_whitespace().collect::<Vec<&str>>();
    if parts.len() > 0 {
        let command = parts[0];
        let args: Vec<&str> = parts.into_iter().skip(1).collect();
        match command {
            "uci" => Ok(Command::Uci),
            "isready" => Ok(Command::IsReady),
            "position" => parse_position(&args),
            "go" => parse_go(&args),
            "stop" => Ok(Command::Stop),
            "quit" | "exit" => Ok(Command::Quit),
            /*"isready" => cmd_isready(cmd, r, s),
            "position" => cmd_position(cmd, r, s),
            "print" => cmd_print(cmd, r, s),
            "go" => cmd_go(cmd, r, s),
            "stop" => cmd_stop(cmd, r, s),
            "ucinewgame" => cmd_newgame(cmd, r, s),
            //"setoption" => cmd_setoption(cmd, r, s),
            "quit" | "exit" => cmd_quit(cmd, r, s),*/
            cmd => { Err("unknown command".to_owned()) }
        }
    }
    else {
        Err("no command".to_owned())
    }
}

fn parse_position(args: &Vec<&str>) -> Result<Command, String> {
    //let position = args[0];
    //args.drain(0..1);
    let mut arg_iter = args.iter();
    let &position = arg_iter.next().ok_or("no position type argument")?;

    let mut board = Board::default();

    if position == "startpos" {
    }
    else if position == "fen" {
        let fen_parts: Vec<&str> = arg_iter.by_ref().take(6).map(|p| *p).collect::<Vec<&str>>();
        if fen_parts.len() != 6 {
            return Err("invalid fen arguments".to_owned());
        }

        board = Board::from_fen(fen_parts.as_slice()).ok_or("invalid fen")?;
    }
    else {
        return Err("invalid position type".to_owned());
    }

    let pmov = arg_iter.next();
    
    
    let moves =
        if let Some(&moves) = pmov {
            if moves != "moves" {
                return Err("expected nothing or 'moves'".to_owned());
            }

            arg_iter.map(|m| String::from(*m)).collect::<Vec<String>>()
        }
        else {
            Vec::new()
        };

    Ok(Command::Position{ pos: board, moves })
}


/*fn run_command(mut cmd: Vec<&str>, r: &Receiver<InterfaceMsg>, s: &Sender<Command>) {
    if cmd.len() > 0 {
        let command = cmd[0];
        cmd.drain(0..1);
        match command {
            "uci" => cmd_uci(cmd),
            "isready" => cmd_isready(cmd, r, s),
            "position" => cmd_position(cmd, r, s),
            "print" => cmd_print(cmd, r, s),
            "go" => cmd_go(cmd, r, s),
            "stop" => cmd_stop(cmd, r, s),
            "ucinewgame" => cmd_newgame(cmd, r, s),
            //"setoption" => cmd_setoption(cmd, r, s),
            "quit" | "exit" => cmd_quit(cmd, r, s),
            cmd => { println!("unknown command: {}", cmd); }
        }
    }
}*/

/*fn cmd_isready(_args: Vec<&str>, _r: &Receiver<InterfaceMsg>, s: &Sender<Command>) {
    s.send(Command::IsReady).unwrap();
}

fn cmd_position(mut args: Vec<&str>, _r: &Receiver<InterfaceMsg>, s: &Sender<Command>) {
    let position = args[0];
    args.drain(0..1);

    let mut game = Board::default();

    if position == "startpos" {
    }
    else if position == "fen" {
        let fen_parts: Vec<&str> = Vec::from(&args[0..6]).into_iter().collect::<Vec<&str>>();
        game = Board::from_fen(fen_parts.as_slice()).unwrap_or_else(|| {
            info!("invalid fen");
            println!("invalid fen");
            Board::default()
        });
        args.drain(0..6);
    }
    let moves = 
        if args.len() > 0 {
            if args[0] != "moves" {
                info!("unexpected {}", args[6]);
                println!("unexpected {}", args[6]);
            }
            args.drain(0..1);


            args.into_iter().map(|m| String::from(m)).collect::<Vec<String>>()
        }
        else {
            Vec::new()
        };

    s.send(Command::Position{ pos: game, moves }).unwrap();
}*/


pub fn parse_go(args: &Vec<&str>) -> Result<Command, String> {
    let mut options: BTreeMap<String, String> = BTreeMap::new();
    let mut opt_name: Option<&str> = None;
    let mut si = SearchInfo::new();

    for &arg in args {
        if let Some(on) = opt_name {
            options.insert(on.to_owned(), arg.to_owned());
            opt_name = None;
        }
        else if arg == "infinite" {
            si.infinite = true;
        }
        else if arg == "perft" {
            si.perft = true;
        }
        else {
            opt_name = Some(arg);
        } 
    }

    si.movetime = options.get("movetime").map(|x| x.parse::<_>().unwrap_or(0));
    si.depth = options.get("depth").map(|x| x.parse::<_>().unwrap_or(0));
    si.wtime = options.get("wtime").map(|x| x.parse::<_>().unwrap_or(0));
    si.btime = options.get("btime").map(|x| x.parse::<_>().unwrap_or(0));
    si.winc = options.get("winc").map(|x| x.parse::<_>().unwrap_or(0));
    si.binc = options.get("binc").map(|x| x.parse::<_>().unwrap_or(0));
    si.movestogo = options.get("movestogo").map(|x| x.parse::<_>().unwrap_or(0));


    Ok(Command::Go(si))
}





